﻿//
//*****************************************************************************
//*                                                                           *
//* COPYRIGHT (c) Maelzel, Inc.,                                              *
//*                                                                           *
//*          Redistribution and use of this code and the algorithms herein    *
//*          contained, with or without modification, is permitted provided   *
//*          that the following conditions are met:                           *
//*          * Redistribution of this code and the algorithms herein          *
//*            contained must retain the above copyright notice, this list    *
//*            of conditions, and the following disclaimer.                   *
//*          * Any use of this code and the algorithms herein contained as    *
//*            text or in binary form must attribute, in a reasonable         *
//*            manner, the provenance and authorship of this code as          *
//*            shown in the copyright above.                                  *
//*                                                                           *
//* DISCLAIMER This code and the algorithms herein contained are provided     *
//*            by the developer ''as is'' without any expressed or implied    *
//*            warranties, including, but not limited to, fitness for any     *
//*            particular purpose or any consequential damage that may        *
//*            occur as a consequence of use or association with this code    *
//*            or the algorithms herein contained.                            *
//*                                                                           *
//* TOP      This file.                                                       *
//* CLASS    StartAgents.cs. (Console-to-GUI interface)                       *
//* AUTHOR   Ovi Chris Rouly                                                  *
//*                                                                           *
//* DATE     11/11/2013                                                       *
//* REVISION X.0.0.8                                                          *
//* BYTES    27,484                                                           *
//* TIME     19:00                                                            *
//* PURPOSE  To facilitate the orderly entry of agents into the model.        *
//* METHOD   By means of software simulation, model proto-humans.             *
//* DETAILS  Created in C# to model proto-humans.                             *
//* ENVIRON  Developed on Intel Pentium D 3.4GHz CPU, 4GB RAM, Nvidia-based   *
//*          PCI-X graphics acceleration.  The OS was MS XP Home SP3.  The    *
//*          Microsoft IDE was Visual Studio 2010, Version 10.0.40219.1       *
//*          SP1 with .NET Framework Version 4.5.50709 SP1.                   *
//*                                                                           *
//*****************************************************************************
//

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;

//
//*****************************************************************************
//

namespace Nursery
{

    //
    //*************************************************************************
    // StartAgents

    public class StartAgents
    {

        //
        //*********************************************************************
        // Class constants, variables, and objects.

        #region Login Sequences

        /* SEED GROUP transmission examples.
           String caller     = <path>             (inserted by system) arg [0].
           String firstName  = "male"               (male=M, female=F) arg [1].
           String lastName   = "G100"               (traces geneology) arg [2].
           String regionName = "region_X005-Y009"        (region name) arg [3].
           String region_x   = "128"              (instantiation locX) arg [4].
           String region_y   = "128"              (instantiation locY) arg [5].
           String age        = "non-zero"         (0=birth, <>0=adult) arg [6].
           This can be read from System.Environment.GetCommandLineArgs ().  The
           assumption is that the agent reads the "age" variable and decides to
           branch accordingly.                                               */

        /* CREATE USER example arguments from XMLRPCHttpClient.cs to OpenSim:
           string password (admin) = "password"
           string user_firstname   = "male"  
           string user_lastname    = "G100"
           string user_password    = "password"
           string user_email       = "none@email.com"
           string start_region_x   = "128"
           string start_region_y   = "128"
           This data format must be configured to XML-RMC, HTTP/1.0.  The
           assumption is that the Created User is in the same region as the
           creating agent.                                                   */

        /* OPENSIM LOGIN example arguments passed to main ( string [] args ):
           string FirstName     = "male"                              args [1].
           string LastName      = "G100"                              args [2].
           string Password      = "password"                          args [3].
           string StartLocation = "uri:region_X005-Y009&4&4&4"        args [4].
           string URI           = "http://10.0.0.12:9000/4/4/4"       args [5].
           This data format was shown good in libOMV 0.9.1.  This data is sent
           to OpenSim from libOMV after it has been prepared.
           For example:
           male G100 password uri:region_X005-Y009&4&4&4 35 where 35 is age  */

        #endregion Login Sequences

        #region Variables Dictionary
        // Constants.
        // for generation zero, the max and min ages of members.
        const int    MIN_AGE = 15;                            // default value.
        const int    MAX_AGE = 35;                            // default value.

        // predetermined cohort center when group 1 is in focus.
        const String GROUP_1_REGION_NAME = "region_X003-Y011";
        const int GROUP_1_REG_X_CNTR = 154;       // approx location in region.
        const int GROUP_1_REG_Y_CNTR = 227;       // approx location in region.
      //const String GROUP_1_REGION_NAME = "region_X000-Y000";
      //const int GROUP_1_REG_X_CNTR = 32;        // approx location in region.
      //const int GROUP_1_REG_Y_CNTR = 32;        // approx location in region.

        // predetermined cohort center when group 2 is in focus.
        const String GROUP_2_REGION_NAME = "region_X005-Y009";
        const int GROUP_2_REG_X_CNTR = 219;//251; // approx location in region.
        const int GROUP_2_REG_Y_CNTR = 187;//160; // approx location in region.
      //const String GROUP_2_REGION_NAME = "region_X000-Y000";
      //const int GROUP_2_REG_X_CNTR = 64;        // approx location in region.
      //const int GROUP_2_REG_Y_CNTR = 64;        // approx location in region.

        // predetermined cohort center when group 3 is in focus.
        const String GROUP_3_REGION_NAME = "region_X012-Y002";
        const int GROUP_3_REG_X_CNTR = 214;       // approx location in region.
        const int GROUP_3_REG_Y_CNTR = 223;       // approx location in region.
      //const String GROUP_3_REGION_NAME = "region_X000-Y000";
      //const int GROUP_3_REG_X_CNTR = 96;        // approx location in region.
      //const int GROUP_3_REG_Y_CNTR = 96;        // approx location in region.

        // predetermined Z axis starting (default) height for all cohorts.
        const int    SEA_LEVEL = 22;           // in meters within the Sim ASL.

        // maximum starting distance from group center regardless of cohort.
        const int LOC_DEVIATION_DIST = 10;  // dist from group & regionXY cntr.

        // miscellaneous.
        const int    LAST_NAME_LENGTH = 4;   // length of an agent's last name.
        const String PASSWORD = "password";                   // default value.
        
        // Variables.
        static String seedGroupCommands = null;     // outputs to command line.
        static String agentFileName = "";  // the file that is the agent model.
        static String agentSubDirectory = "";  // the subdirectory where it is.
        static String nurseryDir = "";         // full path to the agent model.

        static int malesToStartPerGroup = 0;
        static int femalesToStartPerGroup = 0;

        // Objects.
        static Random coin = null;
        static ArtificialGenetics ga = null;

        #endregion Variables Dictionary
        
        //
        //*********************************************************************
        //

        #region Constructor

        /// <summary>
        /// this is the constructor.
        /// </summary>
        public StartAgents ( String newFileName, String newSubDir, int rng )
        {
            agentFileName = newFileName;
            agentSubDirectory = newSubDir;

            nurseryDir = Directory.GetCurrentDirectory () + agentSubDirectory;
            coin = new Random(rng);
            ga = new ArtificialGenetics(rng);
        
        }

        #endregion Constructor
        
        //
        //*********************************************************************
        //

        #region Methods
        
        /// <summary>
        /// this builds and returns a list of Agents that will exist as the
        /// equivalent of one cohort (or one spatially affiliated group).
        /// These agents are assigned names sequentially.  For example,
        /// "male 0001", "male 0002", "male 0003", ..., or ... "female 0099".
        /// </summary>
        public static List < AgentTemplate > BuildOneCohort ( int males,
                                      int females, int randoms, int groups )
        {
            List<AgentTemplate> theAgentList = new List<AgentTemplate> ();

            // user may have asked to have sexes randomly assigned.  This
            // process preassigns values to Set(Fe-)MalesToStartPerGroups ().
            GenerateRandomSexAssignments ( randoms );

            SetMalesToStartPerGroup      // use new and any random male totals.
                ( males + GetMalesToStartPerGroup () );
            SetFemalesToStartPerGroup  // use new and any random female totals.
                ( females + GetFemalesToStartPerGroup () );

            String paddedArray = "";       // this will become their last name.
            String nameString = "";        // this will become their full name.

            // build the pure male part of the list starting with #1.
            for ( int m = 1; m <= GetMalesToStartPerGroup (); m++ )
            {
                nameString = "male ";  // NOTE the trailing space.
                paddedArray = m.ToString ();
                nameString += paddedArray.PadLeft
                    ( LAST_NAME_LENGTH, Convert.ToChar ( "0" ) );
                theAgentList.Add ( new AgentTemplate ( nameString ) );
            }

            // build the pure female part of the list starting with #1.
            for ( int f = 1; f <= GetFemalesToStartPerGroup (); f++ )
            {
                nameString = "female ";  // NOTE the trailing space.
                paddedArray = f.ToString ();
                nameString += paddedArray.PadLeft
                    ( LAST_NAME_LENGTH, Convert.ToChar ( "0" ) );
                theAgentList.Add ( new AgentTemplate ( nameString ) );
            }

            return theAgentList;
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// WARNING: this is a long method.
        /// create process thread for an adult agent using stored data.  Two
        /// types of starting threads are possible, both are supported.  The
        /// first is an OS command process the second a Midwife String var.
        /// </summary>
        public static List < AgentTemplate > CreateGenerationZero
                                           ( List<AgentTemplate> agentList )
        {
            int xDeviation = 0;
            int yDeviation = 0;

            String firstName = "";                         // male=M, female=F.
            String lastName = "";                          // traces geneology.
            String regionName = "";                             // region name.
            int    region_x = 0;                         // instantiation locX.
            int    region_y = 0;                         // instantiation locY.
            String sex = "";                                     // explicitly.

            // outputList contains current set of agents and group identifiers.
            string agentName = "";                          // a working dummy.
            foreach ( AgentTemplate agent in agentList )
            {
                agentName = agent.GetName ();
                if ( agentName.Contains ( "female" ) )
                {
                    firstName = agentName.Substring ( 0, 6 );   // six letters.
                    lastName = agentName.Substring ( 7, 4 );   // four letters.
                    sex = "F";
                }
                else if ( agentName.Contains ( "male" ) )
                {
                    firstName = agentName.Substring ( 0, 4 );  // four letters.
                    lastName =  agentName.Substring ( 5, 4 );  // four letters.
                    sex = "M";
                }
                else
                    Console.WriteLine ( "ERROR in CreateGeneration names" );

                if ( lastName.Substring ( 1, 1 ).Equals( "1" ) )
                {
                    regionName = GROUP_1_REGION_NAME;
                    region_x = GROUP_1_REG_X_CNTR;
                    region_y = GROUP_1_REG_Y_CNTR;
                }
                else if ( lastName.Substring ( 1, 1 ).Equals ( "2" ) )
                {
                    regionName = GROUP_2_REGION_NAME;
                    region_x = GROUP_2_REG_X_CNTR;
                    region_y = GROUP_2_REG_Y_CNTR;
                }
                else if ( lastName.Substring ( 1, 1 ).Equals ( "3" ) )
                {
                    regionName = GROUP_3_REGION_NAME;
                    region_x = GROUP_3_REG_X_CNTR;
                    region_y = GROUP_3_REG_Y_CNTR;
                }
                else
                    Console.WriteLine ( "ERROR in CreateGeneration groups" );

                // develop X.
                xDeviation = coin.Next ( LOC_DEVIATION_DIST );   // 0 to limit.
                xDeviation *= ( ( ( coin.NextDouble () > 0.5D ) ? // x 1 or -1.
                                    true : false ) ?
                                    +1 : -1 );
                region_x += xDeviation;
                if ( region_x < 0 )
                    region_x = 0;
                if ( region_x > 255 )
                    region_x = 255;

                // develop Y.
                yDeviation = coin.Next ( LOC_DEVIATION_DIST );
                yDeviation *= ( ( ( coin.NextDouble () > 0.5D ) ?
                                    true : false ) ?
                                    +1 : -1 );
                region_y += yDeviation;
                if ( region_y < 0 )
                    region_y = 0;
                if ( region_y > 255 )
                    region_y = 255;

                // develop the first part of the login input string.
                seedGroupCommands = firstName + " " + lastName + " " +
                                    PASSWORD + " " +
                                    "uri:" + regionName +
                                    "&" + region_x.ToString () +
                                    "&" + region_y.ToString () +
                                    "&" + SEA_LEVEL.ToString () + " " +
                                    agent.GetAge().ToString() + " " +
                                    ga.makeChromosome(sex);

                // this will start a newbie as on OS command process.
                agent.SetProcess ( CreateNewAgentProcess ( seedGroupCommands,
                                                           nurseryDir,
                                                           agentFileName ) );

                // this will start a newbie as a Midwife command message.
                agent.SetSeedGroupCommands ( seedGroupCommands );

            }

            return agentList;
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this method adds Process (Agent) Start up Information to a Process
        /// Start Information header, starts a new (Agent) Process, then
        /// returns the new process.
        /// </summary>
        private static Process CreateNewAgentProcess ( String newbieDetails,
                                                       String startupFolder,
                                                    String startupFileName )
        {
            const bool USE_NO_WINDOW = false;
            const bool USE_OS_SHELL = true;

            Process p = new Process ();     // this will be the new OS process.
            ProcessStartInfo psi = new ProcessStartInfo ();

            // setup the process start environment.
            psi.CreateNoWindow = USE_NO_WINDOW;
            psi.UseShellExecute = USE_OS_SHELL;
            psi.WindowStyle = ProcessWindowStyle.Normal;
            // set the process start specifics.
            psi.Arguments = newbieDetails;             // describes this agent.
            psi.WorkingDirectory = startupFolder;  // folder where this starts.
            psi.FileName = startupFileName;        // the name of this process.
            // transfer the startup conditions to the process.
            p.StartInfo = psi;

            psi = null;   // give startup conditions memory back to the system.

            return p; // return pointer information to the caller for tracking.

        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// increase the number of groups by a factor of cohort (a # nr.) and
        /// assert the nascent group affiliation into the last names.  Use the
        /// first two character position in the last name for encoding.  For
        /// example: "male G101", "male G102", "male G509", or "female G299".
        /// </summary>
        public static List < AgentTemplate > DuplicateTheCohort ( int groups,
                                             List<AgentTemplate> agentList )
        {
            // capture the original number of agents in the first cohort.
            int originalLength = agentList.Count;

            // this will be the output list.
            List < AgentTemplate > outputList = new List < AgentTemplate > ();
            AgentTemplate agent = null;    // a placeholder from which to work.

            // duplicate first cohort, as a list, "group" times; no changes.
            for ( int group = 0; group < groups; group++ )
                foreach ( AgentTemplate dummy in agentList )
                {
                    agent = new AgentTemplate ();
                    agent.SetName ( dummy.GetName () );
                    agent.SetDisplayLine ( dummy.GetDisplayLine () );
                    agent.SetAge ( dummy.GetAge () );
                    agent.SetProcess ( dummy.GetProcess () );
                    outputList.Add ( agent );
                }

            // insert group affiliation encodings into member's last names.
            String fullName = "";  // work from rght to lft with target string.
            int posFromRght = 0;     // a place to insert "G" and group number.
            int memPntr = 0;          // points to each of the members in turn.
            for ( int cohort = 1; cohort < ( groups + 1 ); cohort++ )
            {
                for ( int member = 0; member < originalLength; member++ )
                {
                    agent = outputList [ memPntr ];
                    fullName = agent.GetName ();

                    // insert a "G" into the last name of the agent.
                    posFromRght = fullName.Length - LAST_NAME_LENGTH;
                    fullName = fullName.Insert ( posFromRght, "G" );
                    fullName = fullName.Remove ( ++posFromRght, 1 );

                    // *ASSUME* groups number 0-9 (as in only one digit).
                    // insert the digit into the last name of the agent.
                    posFromRght = fullName.Length - ( LAST_NAME_LENGTH - 1 );
                    fullName = fullName.Insert ( posFromRght,
                                                        cohort.ToString () );
                    fullName = fullName.Remove ( ++posFromRght, 1 );

                    agent.SetName ( fullName );
                    outputList [ memPntr ] = agent;

                    memPntr++;
                }
            }
            agentList = outputList;                    // transfer the pointer.
            outputList = null;              // return the memory to the System.

            return agentList;
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// randomly assign agent sex values according to parameter limit.
        /// This uses a helper RandomSexAssignment ().
        /// </summary>
        public static void GenerateRandomSexAssignments ( int limit )
        {
            for ( int i = 0; i < limit; i++ )
                RandomSexAssignment ();
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this method returns an age, a number, between min and max
        /// inclusive.
        /// </summary>
        private static int RandomAge ( )
        {
            return coin.Next ( MIN_AGE, ( MAX_AGE + 1 ) );
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// randomly assign an agent a sex value then update the corresponding
        /// sex's total count.
        /// </summary>
        public static void RandomSexAssignment()
        {
            int count = 0;

            if ( coin.NextDouble () > 0.50D )
            {
                count = 1 + GetMalesToStartPerGroup ();
                SetMalesToStartPerGroup ( count );
            }
            else
            {
                count = 1 + GetFemalesToStartPerGroup ();
                SetFemalesToStartPerGroup ( count );
            }
        }

        #endregion Methods

        //
        //*********************************************************************
        //

        #region Getters & Setters

        /// <summary>
        /// this method adds a displayable line saying its group affiliation.
        /// </summary>
        public static List<AgentTemplate> SetDisplayableLines ( int groups,
                                             List<AgentTemplate> agentList )
        {
            // calculate the original length of the first cohort.
            int originalLength = agentList.Count / groups;

            // loop and replicate the encoding of spatial cohort identifier
            // ("group") within each member agent.
            AgentTemplate agent = null;    // a placeholder from which to work.
            String displayableString = "";       // the string to be displayed.
            int memPntr = 0;          // points to each of the members in turn.
            for ( int cohort = 0; cohort < groups; cohort++ )
            {
                for ( int member = 0; member < originalLength; member++ )
                {
                    agent = agentList [ memPntr ];
                    displayableString = agent.GetName () +
                        "\t" + "group nr. " + ( cohort + 1 ).ToString ();
                    agent.SetDisplayLine ( displayableString );
                    agentList [ memPntr ] = agent;
                    memPntr++;
                }          
            }

            return agentList;
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this gives the agents their ages based on CONSTANT constraints.
        /// </summary>
        public static List<AgentTemplate> SetAgentAges
                                           ( List<AgentTemplate> agentList )
        {
            foreach ( AgentTemplate agent in agentList )
                agent.SetAge ( RandomAge () );

            return agentList;
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// gets the count of females variable.
        /// </summary>
        /// <returns>int count of females needed to start</returns>
        public static int GetFemalesToStartPerGroup ()
        {
            return femalesToStartPerGroup;
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// sets the count of females variable.
        /// </summary>
        public static void SetFemalesToStartPerGroup ( int newCount )
        {
            femalesToStartPerGroup = newCount;
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// gets the count of males variable.
        /// </summary>
        /// <returns>int count of males needed to start</returns>
        public static int GetMalesToStartPerGroup ()
        {
            return malesToStartPerGroup;
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// sets the count of males variable.
        /// </summary>
        public static void SetMalesToStartPerGroup ( int newCount )
        {
            malesToStartPerGroup = newCount;
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this method does something.
        /// </summary>
        private void Template ()
        {
            // add code here.
        }

        #endregion Getters & Setters

        //
        //*********************************************************************
        //

    }
}
