﻿//
//*****************************************************************************
//*                                                                           *
//* COPYRIGHT (c) Maelzel, Inc.,                                              *
//*                                                                           *
//*          Redistribution and use of this code and the algorithms herein    *
//*          contained, with or without modification, is permitted provided   *
//*          that the following conditions are met:                           *
//*          * Redistribution of this code and the algorithms herein          *
//*            contained must retain the above copyright notice, this list    *
//*            of conditions, and the following disclaimer.                   *
//*          * Any use of this code and the algorithms herein contained as    *
//*            text or in binary form must attribute, in a reasonable         *
//*            manner, the provenance and authorship of this code as          *
//*            shown in the copyright above.                                  *
//*                                                                           *
//* DISCLAIMER This code and the algorithms herein contained are provided     *
//*            by the developer ''as is'' without any expressed or implied    *
//*            warranties, including, but not limited to, fitness for any     *
//*            particular purpose or any consequential damage that may        *
//*            occur as a consequence of use or association with this code    *
//*            or the algorithms herein contained.                            *
//*                                                                           *
//* TOP      This file.                                                       *
//* CLASS    TopForm.cs. (Top level GUI - MS generated)                       *
//* AUTHOR   Ovi Chris Rouly                                                  *
//*                                                                           *
//* DATE     11/17/2013                                                       *
//* REVISION X.0.0.8                                                          *
//* BYTES    22,860                                                           *
//* TIME     19:00                                                            *
//* PURPOSE  To facilitate the orderly entry of agents into the model.        *
//* METHOD   By means of software simulation, model proto-humans.             *
//* DETAILS  Created in C# to model proto-humans.  Agents can be added to     *
//*          the model in groups of males and females.                        *
//* ENVIRON  Developed on a 64-bit Intel Core2Quad, 2.33GHz, with 4GB RAM.    *
//*          The Zotac m-ITX motherboard had board-based Nvidia 9300 graphics *
//*          acceleration. The Operating System was Microsoft (MS)            *
//*          Windows 8 Pro.  The MS Integrated Development Environment (IDE)  *
//*          MS Visual Studio 2010 Version 10.0.40219.1 SP1Rel and MS .NET    *
//*          Framework Version 4.0.30319 SP1Rel.                              *
//*                                                                           *
//*****************************************************************************
//

using System;
using System.IO;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Net;
using System.Net.Sockets;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using System.Windows.Forms;

using Nursery;

//
//*****************************************************************************
//

namespace Nursery
{

    //
    //*************************************************************************
    // TopForm

    public partial class TopForm : Form
    {

        //
        //*********************************************************************
        // Class constants, variables, and objects.

        #region Login Sequences

        /* SEED GROUP transmission examples.
           String caller     = <path>             (inserted by system) arg [0].
           String firstName  = "male"               (male=M, female=F) arg [1].
           String lastName   = "G100"               (traces geneology) arg [2].
           String regionName = "region_X005-Y009"        (region name) arg [3].
           String region_x   = "128"              (instantiation locX) arg [4].
           String region_y   = "128"              (instantiation locY) arg [5].
           String age        = "non-zero"         (0=birth, <>0=adult) arg [6].
           This can be read from System.Environment.GetCommandLineArgs ().  The
           assumption is that the agent reads the "age" variable and decides to
           branch accordingly.                                               */

        /* CREATE USER example arguments from XMLRPCHttpClient.cs to OpenSim:
           string password (admin) = "password"
           string user_firstname   = "male"  
           string user_lastname    = "G100"
           string user_password    = "password"
           string user_email       = "none@email.com"
           string start_region_x   = "128"
           string start_region_y   = "128"
           This data format must be configured to XML-RMC, HTTP/1.0.  The
           assumption is that the Created User is in the same region as the
           creating agent.                                                   */

        /* OPENSIM LOGIN example arguments passed to main ( string [] args ):
           string FirstName     = "male"                              args [1].
           string LastName      = "G100"                              args [2].
           string Password      = "password"                          args [3].
           string StartLocation = "uri:region_X005-Y009&4&4&4"        args [4].
           string URI           = "http://10.0.0.12:9000/4/4/4"       args [5].
           This data format was shown good in libOMV 0.9.1.  This data is sent
           to OpenSim from libOMV after it has been prepared.                 */

        /* Example: male G100 password uri:region_X000-Y000&128&128&128 35
        */

        #endregion Login Sequences

        #region Variables Dictionary
        // Constants.
        const String AGENT_START_MODE = "net";          // alternative is "OS".
        //const String AGENT_START_MODE = "OS";          // alternative is "net".
        const String AGENT_FILENAME = "Agent.exe";    // agent model file name.
        const String AGENT_SUBDIRECTORY = "\\Offspring\\"; // it sub-directory.
        const int LOGIN_PAUSE = 100; // in ms, simple loop, used with DoEvents.
        const int LOGIN_DELAY = 60;    // in seconds, used w/buttonStart_Click.

        // Variables.

        // Objects.
        public static List<AgentTemplate> agentList = null;
        static int rngSeed = DateTime.Today.Millisecond;
        #endregion Variables Dictionary

        //
        //*********************************************************************
        //

        #region Constructor

        /// <summary>
        /// this is the TopForm constructor.
        /// </summary>
        public TopForm ()
        {
            InitializeComponent ();

            StartAgents tools = new StartAgents ( AGENT_FILENAME,
                                                  AGENT_SUBDIRECTORY,
                                                  rngSeed);

            #region reset form controls
            this.buttonActivate.Enabled = false;
            this.buttonRemoveAgents.Enabled = false;
            this.labelActiveAgentCount.Enabled = false;
            #endregion reset form controls
        }

        #endregion Constructor

        //
        //*********************************************************************
        //

        #region Methods

        /// <summary>
        /// create a new list of agents, equivalent to a cohort whose spatial
        /// affiliations can be identified visually by a naming convention
        /// that incorporates an association or "group" number in the agent's
        /// last name.  This should appear along with a sex and numeric ID.
        /// </summary>
        private void buttonCreate_Click ( object sender, EventArgs e )
        {
            agentList = null;  // start by clearing memory (used or otherwise).
            this.listBoxActiveAgents.Items.Clear ();      // empty the listBox.
            StartAgents.SetMalesToStartPerGroup ( 0 );      // clear the total.
            StartAgents.SetFemalesToStartPerGroup ( 0 );    // clear the total.

            // read from the Form controls the user's requested values.
            int cohorts = ( int ) this.numericUpDownGroups.Value;
            int randoms = ( int ) this.numericUpDownRandomSexSelector.Value;
            int males = ( int ) this.numericUpDownMaleSelector.Value;
            int females = ( int ) this.numericUpDownFemaleSelector.Value;

            if ( randoms == 0 && males == 0 && females == 0 ) return; // leave.

            // create a cohort of agents named "male 00xx" and "female 00xx".
            agentList = StartAgents.BuildOneCohort ( males, females,
                                                    randoms, cohorts );

            // all agents must be between a min and max age set by CONSTANT.
            agentList = StartAgents.SetAgentAges ( agentList );

            // increase the number of groups by a factor of cohort and
            // assert the nascent group affiliation through the last names.
            // For example: "male G101", "female G299".
            agentList = StartAgents.DuplicateTheCohort ( cohorts, agentList );

            // give each member of each cohort displayable line.  For example:
            // "male G501    group nr. 5" and "female G399    group nr. 3"
            agentList = StartAgents.SetDisplayableLines ( cohorts, agentList );

            // create full Processes ready to start and assign them to agents.
            agentList = StartAgents.CreateGenerationZero ( agentList );

            foreach ( AgentTemplate agent in agentList )
            {
                // update the list box and transfer all agent data and
                // processes into the GLOBAL container for user access.
                UpdateAgentListBox ( agent.GetDisplayLine () );
            }
            UpdateAgentCounts (); // reports to GUI tray the count of agents.

            #region reset controls
            this.numericUpDownGroups.Enabled = false;
            this.numericUpDownRandomSexSelector.Enabled = false;
            this.numericUpDownMaleSelector.Enabled = false;
            this.numericUpDownFemaleSelector.Enabled = false;
            this.buttonCreate.Enabled = false;
            this.buttonActivate.Enabled = true;
            this.buttonRemoveAgents.Enabled = true;
            this.labelActiveAgentCount.Enabled = true;
            #endregion equivalent to "setup completed
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// deactivate agents and remove them from the list.
        /// </summary>
        private void buttonRemoveAgents_Click ( object sender, EventArgs e )
        {
            if ( this.listBoxActiveAgents.CausesValidation )
            {
                for ( int i = this.listBoxActiveAgents.Items.Count - 1;
                      i >= 0; i-- )
                {
                    if ( this.listBoxActiveAgents.GetSelected ( i ) )
                    {
                        // point to the item and remove all trace of it.
                        agentList.RemoveAt ( i );
                        this.listBoxActiveAgents.Items.RemoveAt ( i );
                    }
                }

                if ( this.listBoxActiveAgents.Items.Count == 0 )
                {
                    #region reset controls
                    this.numericUpDownGroups.Enabled = true;
                    this.numericUpDownRandomSexSelector.Enabled = true;
                    this.numericUpDownMaleSelector.Enabled = true;
                    this.numericUpDownFemaleSelector.Enabled = true;
                    this.buttonCreate.Enabled = true;
                    this.buttonActivate.Enabled = false;
                    this.buttonRemoveAgents.Enabled = false;
                    this.labelActiveAgentCount.Enabled = false;
                    this.numericUpDownGroups.Value = 1;
                    this.numericUpDownMaleSelector.Value = 0;
                    this.numericUpDownFemaleSelector.Value = 0;
                    this.numericUpDownRandomSexSelector.Value = 0;
                    #endregion reset controls
                }

                UpdateAgentCounts ();   // report to GUI tray counts of agents.
            }
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// the user has chosen to exit.
        /// </summary>
        private void buttonExit_Click ( object sender, EventArgs e )
        {
            if ( ValidateUserEntry () )
            {
                this.Dispose ();
                this.Close ();
                Application.Exit ();
            }
        }

        //
        //*********************************************************************
        //

        /// </summary>
        /// using data specified by the user for startup, pass the data to the
        /// agent, wait for LOGIN_PAUSE time in between logins, then exit.
        /// Let CONSTANT steered startup vector decide to use OS or network
        /// start mode transactions.
        /// </summary>
        private void buttonStart_Click ( object sender, EventArgs e )
        {
            // used solely with the net interface.
            //const String BIRTH_DEMANDED = "instantiateWithPrejudice";
            const String BIRTH_REQUESTED = "instantiationRequest";
            const String ACKNOWLEGEMENT = "understoodProceed";
            const String SAY_PARAMETER_LIST = "parameter list#";

            if (RequestUserWait())
            {
                #region reset controls
                this.buttonActivate.Enabled = false;
                this.buttonRemoveAgents.Enabled = false;
                this.buttonQuit.Enabled = false;
                #endregion reset controls

                MsgInterface msgInterface = new MsgInterface();

                int loginPauses = LOGIN_DELAY * 10;              // in seconds.
                foreach ( AgentTemplate agent in agentList )
                {

                    if (AGENT_START_MODE.Equals("net"))
                    {
                        #region network transactions

                        msgInterface.StartClient();        // start the socket.

                        //Console.WriteLine("sending: " + BIRTH_REQUESTED);
                        Boolean state = msgInterface.Send(BIRTH_REQUESTED);
                        String message = msgInterface.Receive();
                        //Console.WriteLine("received: " + message);
                        // if transmission was OK then check the answer and TX.
                        if (state && (message != null) &&
                            ACKNOWLEGEMENT.Equals(message))
                        {
                            message = agent.GetSeedGroupCommands();
                            message = SAY_PARAMETER_LIST + message;
                            //Console.WriteLine("sending: " + message);
                            state = msgInterface.Send(message);
                        }
                        else
                        {
                            message = "abort";
                            state = msgInterface.Send(message);
                        }

                        msgInterface.StopClient();          // stop the socket.

                        #endregion network transactions
                    }
                    else if ( AGENT_START_MODE.Equals ( "OS" ) )
                    {
                        #region OS command line transactions

                        agent.GetProcess ().Start ();// pop, newbie instantiated.

                        #endregion OS command line transactions
                    }

                    #region pause controls
                    for ( int i = 0; i < ( loginPauses ); i++ )
                    {
                        // currently, each LOGIN_PAUSE is 100 ms.
                        System.Threading.Thread.Sleep ( LOGIN_PAUSE );
                        // having disallowed all programmatic interruptions
                        // allow the Events Scheduler to run in the background.
                        Application.DoEvents ();
                    }
                    #endregion pause controls

                }

                msgInterface = null;            // return memory to OS.
          
                #region reset controls
                this.buttonActivate.Enabled = false;
                this.buttonRemoveAgents.Enabled = false;
                #endregion reset controls

                this.Dispose ();
                this.Close ();
            }

        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this control helps the up down counter.
        /// </summary>
        /// <param name="sender">undefined object</param>
        private void numericUpDownFemaleSelector_ValueChanged
                                              ( object sender, EventArgs e )
        {
            // built-in.
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this control helps the up down counter.
        /// </summary>
        /// <param name="sender">undefined object</param>
        private void numericUpDownGroups_ValueChanged
                                              ( object sender, EventArgs e )
        {
            // built-in.
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this control helps the up down counter.
        /// </summary>
        /// <param name="sender">undefined object</param>
        private void numericUpDownMaleSelector_ValueChanged
                                              ( object sender, EventArgs e )
        {
            // built-in.
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this control helps the up down counter.
        /// </summary>
        /// <param name="sender">undefined object</param>
        private void numericUpDownRandomSexSelector_ValueChanged
                                              ( object sender, EventArgs e )
        {
            // built-in.
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// from MSDN: a message box method template used to ask user to wait
        /// </summary>
        private static bool RequestUserWait ()
        {
            string message = "OK to take " + LOGIN_DELAY +
                             " seconds to start each Agent?";
            string caption = "Initial Virtual World Insertion";
            MessageBoxButtons buttons = MessageBoxButtons.YesNo;
            DialogResult result;

            // Displays the MessageBox.
            result = MessageBox.Show ( message, caption, buttons );

            if ( result == System.Windows.Forms.DialogResult.Yes )
                return true;
            else
                return false;
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this control helps the GUI load.
        /// </summary>
        private void TopForm_Load ( object sender, EventArgs e )
        {
            // built-in.
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// updates the on-screen status bar directly.
        /// </summary>
        private void UpdateAgentCounts ()
        {
            
            int activeAgentNr = this.listBoxActiveAgents.Items.Count;
            this.labelActiveAgentCount.Text = "Active agents:  " +
                                                     activeAgentNr.ToString ();
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// loads the activeAgents listBox.
        /// </summary>
        private void UpdateAgentListBox ( String agentName )
        {   
            this.listBoxActiveAgents.Items.Add ( agentName ); // load listBox.
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// from MSDN: a message box method template used to check user intent
        /// </summary>
        private static bool ValidateUserEntry ()
        {
            string message = "Are sure you want to Exit?";
            string caption = "Confirm Exit Request";
            MessageBoxButtons buttons = MessageBoxButtons.YesNo;
            DialogResult result;

            // Displays the MessageBox.
            result = MessageBox.Show ( message, caption, buttons );

            if ( result == System.Windows.Forms.DialogResult.Yes )
                return true;
            else
                return false;
        }

        //
        //*********************************************************************
        //

        /// <summary>
        /// this method does something.
        /// </summary>
        private void Template ()
        {
            // add code here.
        }

        #endregion Methods

        //
        //*********************************************************************
        //

    }
}
